from datetime import datetime
import os, uuid

from traducteur.lib.model import BaseSQLModel
from traducteur.lib.manager import SQLModelManager

class BaseSQLiteModel(BaseSQLModel):
    id: str = uuid.uuid4().hex

    @property
    def _manager(self):
        return self.__class__.__get_manager()

    @property
    def _col_name(self):
        return self.__class__.__name__

    class Config:
        arbitrary_types_allowed = True
        allow_population_by_field_name = True
    
    def save(self):
        if self.created_at == None:
            self.created_at = datetime.utcnow()
            self.updated_at = datetime.utcnow()
            return self._manager.insert_one(self)
        else:
            self.updated_at = datetime.utcnow()
            return self._manager.update_one(self)

    def delete(self):
        self.deleted_at = datetime.utcnow()
        return self._manager.delete_one(self)

    @classmethod
    def __get_manager(cls):
        con_str = os.environ['TRADUCTEUR_CONNECTION_STR']
        db_name = os.environ['TRADUCTEUR_DATABASE']
        return SQLModelManager(con_str, db_name)

    @classmethod
    def get(cls, id: str):
        manager = cls.__get_manager()
        result = manager.get_one(cls.__name__, id)
        return cls.from_dict(result)

    @classmethod
    def from_dict(cls, values: dict):
        return cls(**values)

    @classmethod
    def create_table(cls):
        schema = cls.schema()
        
        manager = cls.__get_manager()
        manager.create_table(schema)



from typing import Optional
from pymongo import MongoClient
import sqlite3

class BaseDatabaseContext(object):
    def __init__(self, connection_string: str, db_name: str):
        self.connection_string = connection_string
        self.db_name = db_name

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        pass

class MongoContext(BaseDatabaseContext):
    def __init__(self, connection_string: str, db_name: str, column_name: Optional[str] = None):
        super().__init__(connection_string, db_name)
        
        self.client = MongoClient(self.connection_string)
        self.db = self.client[self.db_name]
        self.col = column_name

    def __enter__(self):
        return self.db[self.col] if self.col else self.db

    def __exit__(self, exc_type, exc_value, traceback):
        self.client.close()

class SQLite3Context(BaseDatabaseContext):
    def __init__(self, connection_string: str, db_name: str):
        super().__init__(connection_string, db_name)

        self.client = sqlite3.connect(self.connection_string)
        
    def __enter__(self):
        return self.client.cursor()

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)
        self.client.commit()
        self.client.close()


from datetime import datetime
from pydantic import Field
from typing import Optional
from bson import ObjectId
import os, logging

from traducteur.lib.model import BaseModel
from traducteur.lib.manager import MongoModelManager

class PydanticObjectId(ObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not ObjectId.is_valid(v):
            raise ValueError('Invalid objectid')
        return ObjectId(v)

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(type='string')


class BaseMongoModel(BaseModel):
    id: Optional[PydanticObjectId] = Field(alias='_id')

    @property
    def _manager(self):
        return self.__class__.__get_manager()

    @property
    def _col_name(self):
        return self.__class__.__name__

    class Config:
        arbitrary_types_allowed = True
        allow_population_by_field_name = True
        json_encoders = {
            ObjectId: str,
            PydanticObjectId: str
        }
    
    def save(self):
        if self.created_at == None:
            self.created_at = datetime.utcnow()
            self.updated_at = datetime.utcnow()
            return self._manager.insert_one(self)
        else:
            self.updated_at = datetime.utcnow()
            return self._manager.update_one(self)

    def delete(self):
        self.deleted_at = datetime.utcnow()
        return self._manager.delete_one(self)

    @classmethod
    def __get_manager(cls):
        con_str = os.environ['TRADUCTEUR_CONNECTION_STR']
        db_name = os.environ['TRADUCTEUR_DATABASE']
        return MongoModelManager(con_str, db_name)

    @classmethod
    def get(cls, id: str):
        manager = cls.__get_manager()
        result = manager.get_one(cls.__name__, id)
        return cls.from_dict(result) if result else None

    @classmethod
    def get_where(cls, query, limit: int = 10, **kwargs):
        manager = cls.__get_manager()
        result = manager.get_many(cls.__name__, query, limit=limit, **kwargs)
        return [cls.from_dict(i) for i in result] if len(result) > 0 else None

    @classmethod
    def all(cls, limit: int = 25, **kwargs):
        manager = cls.__get_manager()
        result = manager.get_all(cls.__name__, limit=limit, **kwargs)
        return [cls.from_dict(i) for i in result] if len(result) > 0 else None

    @classmethod
    def exists(cls, id: str) -> bool:
        try:
            return cls.__get_manager().exists(cls.__name__, id=id)
        except Exception:
            return False

    @classmethod
    def exists_where(cls, query) -> bool:
        try:
            return cls.__get_manager().exists_where(cls.__name__, query)
        except Exception as e:
            logging.error(e)
            return False

    @classmethod
    def from_dict(cls, values: dict):
        return cls(**values)


from traducteur.lib.model import BaseModel, BaseSQLModel
from traducteur.lib.context import MongoContext, SQLite3Context
from bson import ObjectId
from pymongo.cursor import Cursor

class BaseModelManager:
    def __init__(self, connection_string: str, database_name):
        self.connection_string = connection_string
        self.database_name = database_name

    def exists(self, col: str, id: str):
        pass

    def exists_where(self, col: str, query):
        pass

    def insert_one(self, item: BaseModel):
        pass

    def get_one(self, col: str, id: str):
        pass

    def get_many(self, col: str, query, limit: int):
        pass

    def update_one(self, update: BaseModel):
        pass

    def delete_one(self, query):
        pass




class SQLModelManager(BaseModelManager):
    def __init__(self, connection_string: str, database_name):
        super().__init__(connection_string, database_name)

    def __match_python_type_sql(self, type_name: str):
        match = {
            'string': 'TEXT',
            'integer': 'INT',
            'float': 'FLOAT(24)',
        }

        if match[type_name] == None:
            raise Exception('No SQL type found for ' + type_name)

        return match[type_name]

    def __check_table(self, name: str):
        query = f"SELECT name FROM sqlite_master WHERE type='table' AND name='{name}';"
        with SQLite3Context(self.connection_string, self.database_name) as db:
            cursor = db.execute(query)
            if cursor.fetchone()[0] == name:
                return True
            return False
        

    def create_table(self, schema: dict):
        if self.__check_table(schema['title']):
            print('Table already exists')
            return True

        cols = []
        for col, val in schema['properties'].items():
            sql_type = self.__match_python_type_sql(val['type'])
            req = f"\tNOT NULL" if col in schema['required'] else ""
            subquery = f"{col}\t{sql_type}{req}"
            cols.append(subquery)

        query = f"CREATE TABLE {schema['title']} ({', '.join(cols)});"
        self.query(query)

    # model actions
    def query(self, query: str):
        with SQLite3Context(self.connection_string, self.database_name) as db:
            return db.execute(query)

    def exists(self, col: str, id: str):
        with SQLite3Context(self.connection_string, self.database_name) as db:
            query = f"SELECT COUNT(id) FROM {col} WHERE id = ?;"
            cursor = db.execute(query, id)
            return len(cursor) >= 1

    def get_one(self, col: str, id: str):
        with SQLite3Context(self.connection_string, self.database_name) as db:
            query = f"SELECT * FROM {col} WHERE id = ?;"
            cursor = db.execute(query, id)
            return cursor.fetchone()

    def get_many(self, col: str, query, limit: int):
        with SQLite3Context(self.connection_string, self.database_name) as db:
            query = f"SELECT * FROM {col} WHERE {', '.join(query)} LIMIT {limit};"
            cursor = db.execute(query)
            return cursor.fetchall()

    def insert_one(self, item: BaseSQLModel):
        with SQLite3Context(self.connection_string, self.database_name) as db:
            query = f"INSERT INTO {item._col_name} ({item.sql_columns}) VALUES ({item.q_marks});"
            print(query)
            cursor = db.execute(query, item.column_values)
            return cursor.lastrowid

    def update_one(self, update: BaseSQLModel):
        with SQLite3Context(self.connection_string, self.database_name) as db:
            query = f"UPDATE {update._col_name} SET {update.sql_update} WHERE id = ?;"
            cursor = db.execute(query, update.id)
            return cursor.lastrowid

    def delete_one(self, item: BaseSQLModel):
        with SQLite3Context(self.connection_string, self.database_name) as db:
            query = f"DELETE FROM {item._col_name} WHERE id = ?;"
            cursor = db.execute(query, item.id)
            return True




class MongoModelManager(BaseModelManager):
    def __init__(self, connection_string: str, database_name):
        super().__init__(connection_string, database_name)

    def _after(self, cursor: Cursor, **kwargs):
        if 'limit' in kwargs:
            cursor.limit(kwargs.get('limit'))

        if 'sort' in kwargs:
            cursor.sort(kwargs.get('sort'))

        return cursor

    def exists(self, col: str, id: str):
        with MongoContext(self.connection_string, self.database_name, col) as column:
            return column.count_documents({'_id': ObjectId(id)}) > 0

    def exists_where(self, col: str, query):
        with MongoContext(self.connection_string, self.database_name, col) as column:
            return column.count_documents(query) > 0

    def get_one(self, col: str, id: str):
        with MongoContext(self.connection_string, self.database_name, col) as column:
            query = {'_id': ObjectId(id)}
            return column.find_one(query)

    def get_all(self, col: str, **kwargs):
        with MongoContext(self.connection_string, self.database_name, col) as column:
            cursor = column.find()
            return list(self._after(cursor, **kwargs))

    def get_many(self, col: str, query, **kwargs):
        with MongoContext(self.connection_string, self.database_name, col) as column:
            cursor = column.find(query)
            return list(self._after(cursor, **kwargs))

    def insert_one(self, item: BaseModel):
        with MongoContext(self.connection_string, self.database_name, item._col_name) as column:
            result = column.insert_one(item.dict())
            item.id = result.inserted_id
            return item

    def update_one(self, update: BaseModel):
        with MongoContext(self.connection_string, self.database_name, update._col_name) as column:
            query = {'_id': ObjectId(update.id) }
            column.update_one(query, {"$set": update.dict()})
            return update

    def delete_one(self, item: BaseModel):
        with MongoContext(self.connection_string, self.database_name, item._col_name) as column:
            query = {'_id': ObjectId(item.id)}
            column.delete_one(query)
            return item


from pydantic import BaseModel as BasePydanticModel
from datetime import datetime
from typing import Optional, Any

class BaseModel(BasePydanticModel):
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    deleted_at: Optional[datetime] = None

    @property
    def column_names(self):
        return [str(col).lower() for col in self.dict().keys()]

    @property
    def column_values(self):
        return tuple([str(v) for v in self.dict().values()])

    def __init__(__pydantic_self__, **data: Any):
        super().__init__(**data)

    def save(self):
        pass

    def delete(self):
        pass

    @classmethod
    def get(cls, id: str):
        pass

    @classmethod
    def get_where(cls, query):
        pass

    @classmethod
    def exists(cls, id):
        pass

    @classmethod
    def exists_where(cls, query):
        pass


class BaseSQLModel(BaseModel):
    @property
    def sql_columns(self):
        return ', '.join(self.column_names)

    @property
    def sql_values(self):
        return ', '.join(self.column_values)

    @property
    def q_marks(self):
        return ', '.join(['?' for _ in range(len(self.column_names))])

    @property
    def sql_update(self):
        return ', '.join([f"{str(col).lower()} = {val}" for col, val in self.dict().items()])


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="traducteur",
    version="0.1.6",
    author="seppedelanghe",
    author_email="seppedelanghe17@gmail.com",
    description="Traducteur is a simple database manager using pydantic models. Currently only supports mongodb.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/seppedelanghe/traducteur",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where='src'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6.1',
)